@@ -23,7 +23,6 @@
 #include "ItemList.h"
 #include "TagMenu.h"
 
-#include <OpSet.h>
 #include <map>
 
 #include <sigc++/sigc++.h>
@@ -32,13 +31,14 @@
 #include "Environment.h"
 
 using namespace std;
+using namespace Tagcoll;
 
 class TagsetCellRenderer : public Gtk::CellRendererText
 {
 protected:
 	ItemList& itemList;
 
-	Glib::Property< OpSet<string> > _property_tagset;
+	Glib::Property< TagSet > _property_tagset;
      
 	/*
 	virtual void get_size_vfunc(Gtk::Widget& widget,
@@ -47,7 +47,7 @@ protected:
 								int* width,    int* height);
 	*/
 
-	virtual void render_vfunc(const Glib::RefPtr<Gdk::Window>& window,
+	virtual void render_vfunc(const Glib::RefPtr<Gdk::Drawable>& drawable,
 								Gtk::Widget& widget,
 								const Gdk::Rectangle& background_area,
 								const Gdk::Rectangle& cell_area,
@@ -65,14 +65,14 @@ public:
 	TagsetCellRenderer(ItemList& itemList);
 	~TagsetCellRenderer() {}
 	
-	Glib::PropertyProxy< OpSet<string> > property_tagset();
+	Glib::PropertyProxy< TagSet > property_tagset();
 };
 
 TagsetCellRenderer::TagsetCellRenderer(ItemList& itemList) :
 	Glib::ObjectBase(typeid (TagsetCellRenderer)),
 	Gtk::CellRendererText(),
 	itemList(itemList),
-	_property_tagset(*this, "tagset", OpSet<string>())
+	_property_tagset(*this, "tagset", TagSet())
 {
 	/*
 	property_mode() = Gtk::CELL_RENDERER_MODE_ACTIVATABLE;
@@ -81,7 +81,7 @@ TagsetCellRenderer::TagsetCellRenderer(ItemList& itemList) :
 	*/
 }
 
-Glib::PropertyProxy< OpSet<string> > TagsetCellRenderer::property_tagset()
+Glib::PropertyProxy< TagSet > TagsetCellRenderer::property_tagset()
 {
 	return _property_tagset.get_proxy();
 }
@@ -121,24 +121,24 @@ void TagsetCellRenderer::get_size_vfunc(Gtk::Widget&,
 }
 */
 
-void TagsetCellRenderer::render_vfunc(const Glib::RefPtr<Gdk::Window>& window,
+void TagsetCellRenderer::render_vfunc(const Glib::RefPtr<Gdk::Drawable>& drawable,
 										Gtk::Widget& widget,
 										const Gdk::Rectangle& rect1,
 										const Gdk::Rectangle& cell_area,
 										const Gdk::Rectangle& rect2,
 										Gtk::CellRendererState flags)
 {
-	OpSet<string> ts = _property_tagset.get_value() - itemList._tagset;
+	TagSet ts = _property_tagset.get_value() - itemList._tagset;
 	string tags;
-	for (OpSet<string>::const_iterator j = ts.begin();
+	for (TagSet::const_iterator j = ts.begin();
 			j != ts.end(); j++)
 		if (j == ts.begin())
-			tags += *j;
+			tags += j->fullname();
 		else
-			tags += ", " + *j;
+			tags += ", " + j->fullname();
 
 	property_text().set_value(tags);
-	Gtk::CellRendererText::render_vfunc(window, widget, rect1, cell_area, rect2, flags);
+	Gtk::CellRendererText::render_vfunc(drawable, widget, rect1, cell_area, rect2, flags);
 }
 
 
@@ -207,59 +207,65 @@ void ItemList::on_document_changed()
 	updateList();
 }
 
-void ItemList::on_selected_row_iterated(const Gtk::TreeModel::iterator& iter)
-{
-	Glib::ustring glibitem = (*iter)[itemListModelColumns.name];
-	selected.push_back(glibitem);
-}
-	
-void ItemList::on_add_tag(string tag)
+void ItemList::on_add_tag(Tag tag)
 {
 	Glib::RefPtr<Gtk::TreeSelection> sel = itemList.get_selection();
 	
 	// Build the list of selected items
 	selected.clear();
-	sel->selected_foreach(
-			    SigC::slot(*this, &ItemList::on_selected_row_iterated));
-	
+
+	Gtk::TreeSelection::ListHandle_Path lhp = sel->get_selected_rows();
+	for (Gtk::TreeSelection::ListHandle_Path::const_iterator i = lhp.begin(); i != lhp.end(); i++)
+	{
+		Gtk::TreeModel::Row row = *itemListModel->get_iter(*i);
+		Glib::ustring s = row[itemListModelColumns.name];
+		selected.push_back(s);
+	}
+
 	// Build the Change
-	TagCollection<int>::Change change;
+	PatchList<string, Tag> change;
 	for (vector<string>::const_iterator i = selected.begin();
 			i != selected.end(); i++)
 	{
-		int handle = doc.handles().getHandle(*i);
-		OpSet<string> ts = doc.collection().getTagsetForItem(handle);
-		change.insert(make_pair(handle, ts + tag));
+		Patch<string, Tag> patch(*i);
+		patch.add(tag);
+		change.addPatch(patch);
 	}
 
 	// Post the change
 	do_signal_request_tagcoll_change(change);
 }
 
-void ItemList::on_remove_tag(string tag)
+void ItemList::on_remove_tag(Tag tag)
 {
 	Glib::RefPtr<Gtk::TreeSelection> sel = itemList.get_selection();
 	
 	// Build the list of selected items
 	selected.clear();
-	sel->selected_foreach(
-			    SigC::slot(*this, &ItemList::on_selected_row_iterated));
+
+	Gtk::TreeSelection::ListHandle_Path lhp = sel->get_selected_rows();
+	for (Gtk::TreeSelection::ListHandle_Path::const_iterator i = lhp.begin(); i != lhp.end(); i++)
+	{
+		Gtk::TreeModel::Row row = *itemListModel->get_iter(*i);
+		Glib::ustring s = row[itemListModelColumns.name];
+		selected.push_back(s);
+	}
 	
 	// Build the Change
-	TagCollection<int>::Change change;
+	PatchList<string, Tag> change;
 	for (vector<string>::const_iterator i = selected.begin();
 			i != selected.end(); i++)
 	{
-		int handle = doc.handles().getHandle(*i);
-		OpSet<string> ts = doc.collection().getTagsetForItem(handle);
-		change.insert(make_pair(handle, ts - tag));
+		Patch<string, Tag> patch(*i);
+		patch.remove(tag);
+		change.addPatch(patch);
 	}
 
 	// Post the change
 	do_signal_request_tagcoll_change(change);
 }
 
-void ItemList::do_signal_request_tagcoll_change(TagCollection<int>::Change change)
+void ItemList::do_signal_request_tagcoll_change(PatchList<string, Tag> change)
 {
 	signal_request_tagcoll_change().emit(change);
 }
@@ -284,13 +290,13 @@ void ItemList::do_signal_request_item_move()
 	signal_request_item_move().emit();
 }
 
-void ItemList::do_signal_select_tagset(OpSet<string> tagset)
+void ItemList::do_signal_select_tagset(TagSet tagset)
 {
 	//debug("SRTA %.*s %.*s\n", PFSTR(item), PFSTR(tag));
 	signal_select_tagset().emit(tagset);
 }
 
-void ItemList::do_signal_select_tagset_other_panel(OpSet<string> tagset)
+void ItemList::do_signal_select_tagset_other_panel(TagSet tagset)
 {
 	//debug("SRTA %.*s %.*s\n", PFSTR(item), PFSTR(tag));
 	signal_select_tagset_other_panel().emit(tagset);
@@ -313,7 +319,7 @@ bool ItemList::on_event(GdkEvent* e)
 
 			Gtk::TreeModel::Row row = *itemListModel->get_iter(path);
 			Glib::ustring tag = row[itemListModelColumns.name];
-			OpSet<string> ts = row[itemListModelColumns.tags];
+			TagSet ts = row[itemListModelColumns.tags];
 			
 			debug("Name: %.*s, %d tags\n", PFSTR(tag), ts.size());
 
@@ -325,40 +331,40 @@ bool ItemList::on_event(GdkEvent* e)
 			if (rows == 1)
 			{
 				itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Select this tag set",
-							SigC::bind< OpSet<string> >(
-								SigC::slot(*this, &ItemList::do_signal_select_tagset), ts)));
+							sigc::bind< TagSet >(
+								sigc::mem_fun(*this, &ItemList::do_signal_select_tagset), ts)));
 				itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Select this tag set in the other panel",
-							SigC::bind< OpSet<string> >(
-								SigC::slot(*this, &ItemList::do_signal_select_tagset_other_panel), ts)));
+							sigc::bind< TagSet >(
+								sigc::mem_fun(*this, &ItemList::do_signal_select_tagset_other_panel), ts)));
 			}
 			else
 			{
 				itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Merge",
-							SigC::slot(*this, &ItemList::do_signal_request_tagset_merge)));
+							sigc::mem_fun(*this, &ItemList::do_signal_request_tagset_merge)));
 				itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Intersect",
-							SigC::slot(*this, &ItemList::do_signal_request_tagset_intersect)));
+							sigc::mem_fun(*this, &ItemList::do_signal_request_tagset_intersect)));
 			}
 			itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Copy to other panel",
-						SigC::slot(*this, &ItemList::do_signal_request_item_copy)));
+						sigc::mem_fun(*this, &ItemList::do_signal_request_item_copy)));
 			itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Move to other panel",
-						SigC::slot(*this, &ItemList::do_signal_request_item_move)));
+						sigc::mem_fun(*this, &ItemList::do_signal_request_item_move)));
 
-			TagMenu* addMenu = new TagMenu();
+			TagMenu<TagcollDocument<string> >* addMenu = new TagMenu<TagcollDocument<string> >();
 			addMenu->set_manage();
 			addMenu->populateUnselected(doc, ts);
 			string stag = tag;
 			addMenu->signal_selected().connect(
-						SigC::slot(*this, &ItemList::on_add_tag));
+						sigc::mem_fun(*this, &ItemList::on_add_tag));
 			itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("_Add...", *addMenu));
 	
 			if (!ts.empty())
 			{
 				itemPopup.items().push_back(Gtk::Menu_Helpers::SeparatorElem());
-				for (OpSet<string>::const_iterator i = ts.begin();
+				for (TagSet::const_iterator i = ts.begin();
 						i != ts.end(); i++)
-					itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Remove " + *i,
-						SigC::bind<string>(
-							SigC::slot(*this, &ItemList::on_remove_tag), *i)));
+					itemPopup.items().push_back(Gtk::Menu_Helpers::MenuElem("Remove " + i->fullname(),
+						sigc::bind<Tag>(
+							sigc::mem_fun(*this, &ItemList::on_remove_tag), *i)));
 			}
 
 			itemPopup.popup(e->button.button, e->button.time);
@@ -399,14 +405,10 @@ ItemList::ItemList(TagcollDocument<std::string>& doc)
 	itemList.append_column("Name", itemListModelColumns.name);
 
 	{
-		TagsetCellRenderer*const renderer = new TagsetCellRenderer(*this);
+		TagsetCellRenderer*const renderer = Gtk::manage(new TagsetCellRenderer(*this));
 		Gtk::TreeViewColumn*const column  = new Gtk::TreeViewColumn("Tags", *Gtk::manage(renderer));
 		itemList.append_column(*Gtk::manage(column));
 		column->add_attribute(renderer->property_tagset(), itemListModelColumns.tags);
-
-		/*
-		renderer->signal_toggled().connect(SigC::slot(*this, &AppWindow::on_cell_toggled));
-		*/
 	}
 	//itemList.append_column("Tags", itemListModelColumns.tags);
 
@@ -415,82 +417,86 @@ ItemList::ItemList(TagcollDocument<std::string>& doc)
 
 	Glib::RefPtr<Gtk::TreeSelection> itemListSelection = itemList.get_selection();
 	itemListSelection->set_mode(Gtk::SELECTION_MULTIPLE);
-	itemListSelection->signal_changed().connect(SigC::slot(*this, &ItemList::do_selection_changed));
+	itemListSelection->signal_changed().connect(sigc::mem_fun(*this, &ItemList::do_selection_changed));
 	
 	itemList.add_events(Gdk::BUTTON_PRESS_MASK);
-	itemList.signal_event().connect(SigC::slot(*this, &ItemList::on_event));
+	itemList.signal_event().connect(sigc::mem_fun(*this, &ItemList::on_event));
 
 	// Setup the lists as a drag source
 	std::list<Gtk::TargetEntry> listTargets;
 	listTargets.push_back(Gtk::TargetEntry("TAGCOLL"));
 	listTargets.push_back(Gtk::TargetEntry("text/plain"));
 	itemList.drag_source_set(listTargets, Gdk::ModifierType(GDK_BUTTON1_MASK), Gdk::ACTION_COPY | Gdk::ACTION_MOVE);
-	itemList.signal_drag_data_get().connect(SigC::slot(*this, &ItemList::on_itemList_drag_data_get));
+	itemList.signal_drag_data_get().connect(sigc::mem_fun(*this, &ItemList::on_itemList_drag_data_get));
 
 	listTargets.clear();
 	listTargets.push_back(Gtk::TargetEntry("TAG"));
 	listTargets.push_back(Gtk::TargetEntry("TAGCOLL"));
 	itemList.drag_dest_set(listTargets, Gtk::DEST_DEFAULT_ALL, Gdk::ACTION_COPY | Gdk::ACTION_MOVE);
-	itemList.signal_drag_data_received().connect(SigC::slot(*this, &ItemList::on_itemList_drop_drag_data_received) );
+	itemList.signal_drag_data_received().connect(sigc::mem_fun(*this, &ItemList::on_itemList_drop_drag_data_received) );
 
 	// Fill the list
 	updateList();
 	
 	/*
-	selectedItemList.signal_row_activated().connect(SigC::slot(*this, &ItemList::on_selectedList_row_activated));
-	availableItemList.signal_row_activated().connect(SigC::slot(*this, &ItemList::on_availableList_row_activated));
+	selectedItemList.signal_row_activated().connect(sigc::slot(*this, &ItemList::on_selectedList_row_activated));
+	availableItemList.signal_row_activated().connect(sigc::slot(*this, &ItemList::on_availableList_row_activated));
 	*/
-	itemList.signal_focus_in_event().connect(SigC::slot(*this, &ItemList::on_focus_in));
-	filterEdit.signal_changed().connect(SigC::slot(*this, &ItemList::on_filter_changed));
-	doc.signal_changed().connect(SigC::slot(*this, &ItemList::on_document_changed));
+	itemList.signal_focus_in_event().connect(sigc::mem_fun(*this, &ItemList::on_focus_in));
+	filterEdit.signal_changed().connect(sigc::mem_fun(*this, &ItemList::on_filter_changed));
+	doc.signal_changed().connect(sigc::mem_fun(*this, &ItemList::on_document_changed));
 }
 
 void ItemList::on_itemList_drag_data_get(
-		const Glib::RefPtr<Gdk::DragContext>&, GtkSelectionData* selection_data, guint, guint)
+		const Glib::RefPtr<Gdk::DragContext>&, Gtk::SelectionData& selection_data, guint, guint)
 {
-	std::map<std::string, OpSet<string> > sel = getSelection();
+	std::map<std::string, TagSet > sel = getSelection();
 	if (sel.size())
 	{
 		Glib::ustring str;
-		for (std::map< string, OpSet<string> >::const_iterator i = sel.begin();
+		for (std::map< string, TagSet >::const_iterator i = sel.begin();
 				i != sel.end(); i++)
 		{
 			str += i->first;
 
-			for (OpSet<string>::const_iterator j = i->second.begin();
+			for (TagSet::const_iterator j = i->second.begin();
 					j != i->second.end(); j++)
 			if (j == i->second.begin())
-				str += ": " + *j;
+				str += ": " + j->fullname();
 			else
-				str += ", " + *j;
+				str += ", " + j->fullname();
 
 			str += "\n";
 		}
 		
-		gtk_selection_data_set(selection_data, selection_data->target, 8, (const guchar*)str.data(), str.size());
+		selection_data.set(selection_data.get_target(), 8, (const guchar*)str.data(), str.size());
 	}
 }
 
 void ItemList::on_itemList_drop_drag_data_received(
-		const Glib::RefPtr<Gdk::DragContext>& context, int, int, GtkSelectionData* selection_data, guint, guint time)
+		const Glib::RefPtr<Gdk::DragContext>& context, int, int, const Gtk::SelectionData& selection_data, guint, guint time)
 {
-	if ((selection_data->length >= 0) && (selection_data->format == 8))
+	if ((selection_data.get_length() >= 0) && (selection_data.get_format() == 8))
 	{
-		GdkAtom atype = selection_data->type;
-		string type = Gdk::AtomStringTraits::to_cpp_type(atype);
+		string type = selection_data.get_data_type();
 		warning("Type: %.*s\n", PFSTR(type));
-		Glib::ustring data((const char*)selection_data->data, selection_data->length);
-		warning("Received: %.*s\n", PFSTR(data));
+		Glib::ustring data((const char*)selection_data.get_data(), selection_data.get_length());
+		Tag tag = doc.vocabulary().findTag(data);
+		warning("Received: %.*s [%sfound in vocabulary]\n", PFSTR(data), tag ? "" : "not ");
 
-		if (type == "TAG")
+		if (type == "TAG" && tag)
 		{
-			std::map< std::string, OpSet<string> > sel = getSelection();
+			std::map< std::string, TagSet > sel = getSelection();
 
-			TagCollection<int>::Change change;
+			PatchList<string, Tag> change;
 			
-			for (std::map< string, OpSet<string> >::const_iterator i = sel.begin();
+			for (std::map< string, TagSet >::const_iterator i = sel.begin();
 					i != sel.end(); i++)
-				change.insert(make_pair(doc.handles().getHandle(i->first), i->second + data));
+			{
+				Patch<string, Tag> patch(i->first);
+				patch.add(tag);
+				change.addPatch(patch);
+			}
 			
 			doc.applyChange(change);
 		}
@@ -508,48 +514,42 @@ void ItemList::updateList()
 {
 	time_t start = time(NULL);
 
-	std::map< std::string, OpSet<string> > sel = getSelection();
+	std::map< std::string, TagSet > sel = getSelection();
 	
 	if (_tagset.empty())
 	{
-		OpSet<int> items = doc.collection().getUntaggedItems();
-
-		// Get a sorted list of items
-		OpSet<string> tags;
-		for (OpSet<int>::const_iterator i = items.begin(); i != items.end(); i++)
-			tags += doc.handles().getItem(*i);
+		// Display untagged items
+		OpSet<string> items = doc.collection().getUntaggedItems();
 
 		Gtk::TreeModel::Row row;
 		itemListModel->clear();
-		for (OpSet<string>::const_iterator i = tags.begin(); i != tags.end(); i++)
+		for (OpSet<string>::const_iterator i = items.begin(); i != items.end(); i++)
 		{
 			if (filter.empty() || i->find(filter) != string::npos)
 			{
 				row = *(itemListModel->append());
 				row[itemListModelColumns.name] = *i;
-				row[itemListModelColumns.tags] = OpSet<string>();
+				row[itemListModelColumns.tags] = TagSet();
 				if (sel.find(*i) != sel.end())
 					itemList.get_selection()->select(row);
 			}
 		}
 	} else {
-		std::map< int, OpSet<string> > items = doc.collection().getCompanionItemsAndTagsets(_tagset);
-
-		// Get a sorted list of items
-		std::map< string, OpSet<string> > names;
-		for (std::map< int, OpSet<string> >::const_iterator i = items.begin(); i != items.end(); i++)
-			names.insert(make_pair(doc.handles().getItem(i->first), i->second));
+		// Display items with the given tags
+		OpSet<string> items = doc.collection().getItems(_tagset);
+		
+		//std::map< string, OpSet<Tag> > items = doc.collection().getCompanionItemsAndTagsets(_tagset);
 
 		Gtk::TreeModel::Row row;
 		itemListModel->clear();
-		for (std::map< string, OpSet<string> >::const_iterator i = names.begin(); i != names.end(); i++)
+		for (OpSet<string>::const_iterator i = items.begin(); i != items.end(); i++)
 		{
-			if (filter.empty() || i->first.find(filter) != string::npos)
+			if (filter.empty() || i->find(filter) != string::npos)
 			{
 				row = *(itemListModel->append());
-				row[itemListModelColumns.name] = i->first;
-				row[itemListModelColumns.tags] = i->second;
-				if (sel.find(i->first) != sel.end())
+				row[itemListModelColumns.name] = *i;
+				row[itemListModelColumns.tags] = doc.collection().getTags(*i);
+				if (sel.find(*i) != sel.end())
 					itemList.get_selection()->select(row);
 			}
 		}
@@ -560,16 +560,16 @@ void ItemList::updateList()
 		fprintf(stderr, "ItemList::updateList: %d seconds\n", (end-start));
 }
 
-map< string, OpSet<string> > ItemList::getAllItems()
+map< string, TagSet > ItemList::getAllItems()
 {
-	std::map< string, OpSet<string> > res;
+	std::map< string, TagSet > res;
 
 	for (Gtk::TreeModel::Children::const_iterator i = itemListModel->children().begin();
 			i != itemListModel->children().end(); i++)
 	{
 		Gtk::TreeModel::Row row = *i;
 		Glib::ustring s = row[itemListModelColumns.name];
-		OpSet<string> ts = row[itemListModelColumns.tags];
+		TagSet ts = row[itemListModelColumns.tags];
 		res.insert(make_pair(s, ts));
 		debug("Found: %.*s\n", PFSTR(s));
 	}
@@ -577,17 +577,17 @@ map< string, OpSet<string> > ItemList::getAllItems()
 	return res;
 }
 
-map< string, OpSet<string> > ItemList::getSelection()
+map< string, TagSet > ItemList::getSelection()
 {
 	Glib::RefPtr<Gtk::TreeSelection> itemSelection = itemList.get_selection();
 	Gtk::TreeSelection::ListHandle_Path lhp = itemSelection->get_selected_rows();
-	std::map< string, OpSet<string> > res;
+	std::map< string, TagSet > res;
 
 	for (Gtk::TreeSelection::ListHandle_Path::const_iterator i = lhp.begin(); i != lhp.end(); i++)
 	{
 		Gtk::TreeModel::Row row = *itemListModel->get_iter(*i);
 		Glib::ustring s = row[itemListModelColumns.name];
-		OpSet<string> ts = row[itemListModelColumns.tags];
+		TagSet ts = row[itemListModelColumns.tags];
 		res.insert(make_pair(s, ts));
 	}
 
@@ -595,7 +595,7 @@ map< string, OpSet<string> > ItemList::getSelection()
 
 	/*
 	ItemListSelectionDataGatherer data(*this);
-	itemSelection->selected_foreach(SigC::slot(data, &ItemListSelectionDataGatherer::selected_row_callback));
+	itemSelection->selected_foreach(sigc::slot(data, &ItemListSelectionDataGatherer::selected_row_callback));
 	*/
 	//return data.data();
 }
